macro_rules! ready {
($return_value:expr) => { ... };
}Expand description
Behaves like core::task::ready!, but uses return_borrowed! rather than return so it can be used within maybe_borrow! or try_maybe_borrow!.
This macro is only available within the aforementioned macros.